/* Copyright (c) 2009 The Regents of the Signet Group.
*  All rights reserved.
*
*  Permission to use, copy, modify, and distribute this software and its
*  documentation for any purpose, without fee, and without written
*  agreement is hereby granted, provided that the above copyright
*  notice, the (updated) modification history and the author appear in
*  all copies of this source code.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS `AS IS'
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
*  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS
*  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, LOSS OF USE, DATA,
*  OR PROFITS) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
*  THE POSSIBILITY OF SUCH DAMAGE.
*/

package net.tinyos.signet;

import java.io.BufferedReader;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.SequenceInputStream;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.Queue;
import java.util.StringTokenizer;
import java.util.concurrent.Semaphore;

public class SecureSynapseInterface {
	private static final int DEFAULT_CONCURRENT_SERIALS=8;
	private static Semaphore serialSem=new Semaphore(DEFAULT_CONCURRENT_SERIALS);	
	private static boolean useEncryption=false,useDoSProtection=false,useAuthentication=false;
	
	private static OutputPrinter sysop=new SystemOutputPrinter();
	
	private interface OutputPrinter{
		public void print(String s);
	}
	
	private static class SystemOutputPrinter implements OutputPrinter{
		@Override
		public void print(String arg0) {
			synchronized(System.out){
				System.out.print(arg0);
			}			
		}
	}
	
	private static abstract class Command{
		private OutputPrinter op=null;
		
		public void setOutputPrinter(OutputPrinter op){
			this.op=op;
		}
		
		public abstract void start() throws Exception;
		public abstract boolean waitFor() throws Exception;
		public boolean execute() throws Exception{
			start();
			return waitFor();
		}
		
		protected void printOut(String s){
			if(op!=null)
				op.print(s);
		}
	}
	
	private static class CommandDecorator extends Command{
		private Command innerCommand;
		
		public CommandDecorator(Command innerCommand){
			this.innerCommand=innerCommand;
		}
		
		@Override
		public void start() throws Exception {
			innerCommand.start();			
		}
		@Override
		public boolean waitFor() throws Exception {
			return innerCommand.waitFor();
		}
		
		@Override
		public void setOutputPrinter(OutputPrinter arg0) {
			super.setOutputPrinter(arg0);
			innerCommand.setOutputPrinter(arg0);
		}
		
		@Override
		public String toString() {
			return innerCommand.toString();
		}
		
		@Override
		protected void printOut(String arg0) {
			innerCommand.printOut(arg0);
		}		
	}
	
	private static class NullCommand extends Command{
		
		@Override
		public void start() throws Exception {			
		}
		@Override
		public boolean waitFor() throws Exception {
			return true;
		}
		@Override
		public String toString() {
			return "Null command";
		}
	}	
	
	private static class RetriableCommand extends CommandDecorator{
		private static final int DEFAULT_RETRY_COUNT=5;
		private int retryCount;
		
		public RetriableCommand(int retryCount,Command innerCommand){
			super(innerCommand);
			this.retryCount=retryCount;			
		}
		
		public RetriableCommand(Command innerCommand){
			this(DEFAULT_RETRY_COUNT,innerCommand);						
		}

		@Override
		public boolean waitFor() throws Exception {
			int retries=0;
			while((!super.waitFor())&&(retries<retryCount)){
				retries++;
				printOut("\n*** Command failed, retry ("+retries+") ***\n\n");
				super.start();
			}
			return (retries<retryCount);	
		}
	}
	
	private static class DelayedCommand extends CommandDecorator{
		private static final long DEFAULT_DELAY=1000;
		private long delay;
		
		public DelayedCommand(long delay,Command innerCommand){
			super(innerCommand);
			this.delay=delay;			
		}
		
		public DelayedCommand(Command innerCommand){
			this(DEFAULT_DELAY,innerCommand);						
		}

		@Override
		public void start() throws Exception {
			printOut("Execution delayed by "+delay+"ms.\n");
			Thread.sleep(delay);
			printOut("Executing now.\n");
			super.start();
		}
	}	
	
	private static class SemaphoredCommand extends CommandDecorator{
		private Semaphore sem;
		
		public SemaphoredCommand(Semaphore sem,Command innerCommand){
			super(innerCommand);
			this.sem=sem;			
		}
		
		@Override
		public void start() throws Exception {
			sem.acquire();
			super.start();
			
		}
		
		@Override
		public boolean waitFor() throws Exception {
			boolean result= super.waitFor();
			sem.release();
			return result;
		}
	}		
	
	private static class InteractiveWaitCommand extends Command{
		private String message;
		
		public InteractiveWaitCommand(String message) {
			this.message=message;
		}
		
		@Override
		public void start() throws Exception {
			System.out.println(message+"\n.Press any key to continue");
		}
		@Override
		public boolean waitFor() throws Exception {
			System.in.read();
			return true;
		}
	}
	
	private static class CommandLineCommand extends Command{
		private String command,args[];
		private File dir;
		private Process proc;
		private Semaphore sem=new Semaphore(0);
		private long timeout=0;
		private Semaphore timeoutSem=null;
		private boolean timeoutExitError=false;
		
		public CommandLineCommand(String command,String args[],File dir){
			this.command=command;
			this.args=args;
			this.dir=dir;			
		}
		
		@Override
		public void start() throws IOException {
			String[] cmdarray=new String[args.length+1];
			cmdarray[0]=command;
			for(int i=0;i<args.length;i++)
				cmdarray[i+1]=args[i];
			proc=Runtime.getRuntime().exec(cmdarray, null, dir);
			new Thread(
					new Runnable(){
						@Override
						public void run() {
							BufferedReader br=new BufferedReader(new InputStreamReader(proc.getInputStream()));
							String s;
							try {
								while((s=br.readLine())!=null){
									printOut(s+"\n");
								}									
							} catch (IOException e) {								
							}
							sem.release();
						}
					}
			).start();
			new Thread(
					new Runnable(){
						@Override
						public void run() {
							BufferedReader br=new BufferedReader(new InputStreamReader(proc.getErrorStream()));
							String s;
							try {
								while((s=br.readLine())!=null){
									printOut(s+"\n");
								}									
							} catch (IOException e) {								
							}
							sem.release();
						}
					}
			).start();			
			if(timeout>0){
				new Thread(new Runnable(){
					@Override
					public void run() {
						try {
							Thread.sleep(timeout);
							proc.destroy();
						} catch (InterruptedException e) {
						}					
					}
				}).start();				
			}else if(timeoutSem!=null){
				new Thread(new Runnable(){
					@Override
					public void run() {
						try {
							timeoutSem.acquire();
							timeoutSem.release();
							proc.destroy();
						} catch (InterruptedException e) {
						}					
					}
				}).start();				
			}
		}

		public void setTimeout(long time,boolean exitError){
			timeout=time;
			timeoutExitError=exitError;
		}
		
		public void setTimeoutSemaphore(Semaphore sem,boolean exitError){
			timeoutSem=sem;
			timeoutExitError=exitError;
		}
		
		@Override
		public boolean waitFor() throws Exception{
			
			boolean result;
			try {
				result = proc.waitFor()==0;
			} catch (InterruptedException e) {
				if(timeout!=0) result=!timeoutExitError;
				else throw e;
			}			
			sem.acquire();
			proc.destroy();
			return result;
		}		
		
			
		@Override
		public String toString() {
			String sargs="";
			for(int i=0;i<args.length;i++)
				sargs+=args[i]+" ";
			return "Command line command \""+command+" "+sargs+"\" executed in dir "+(dir!=null?dir.toString():"");
		}
		
	}
	

	
	
	private static class Node{
		public String name,port,platform;
		
		public Node(String name,String port,String platform){
			this.name=name;
			this.port=port;
			this.platform=platform;
		}
		
		@Override
		public String toString() {
			return "Node "+name+" ("+platform+") on port "+port;
		}
	}
	
	private static class Network{
		private Node bs=null;
		private ArrayList<Node> nodes=new ArrayList<Node>();
		
		public void addNode(Node node){
			bs=null;
			nodes.add(node);
		}
		
		public ArrayList<Node> getNodes(){
			return new ArrayList<Node>(nodes);
		}
		
		public Node getBaseStation(){
			if(bs==null){
				for (Node node : nodes) {
					if(bs==null)
						bs=node;
					else{
						if(node.name.compareTo(bs.name)<0)
							bs=node;
					}						
				}
			}
			return bs;
		}
		
		public Node findNodeByPort(String port){
			for (Node node : nodes) {
				if(node.port.equals(port))
					return node;						
			}			
			return null;
		}
		
		@Override
		public String toString() {
			String result="";
			for (Node node : nodes) {
				result+=node.toString()+"\n";
			}
			return result;
		}
	}
	
	private static class NodeDiscovery{
		private String motelistout="";
		public Network discover(){
			Network result=new Network();
			String args[]={"-c"};
			CommandLineCommand cmd=new CommandLineCommand("motelist",args,null);
			cmd.setOutputPrinter(new OutputPrinter(){
				public void print(String arg0) {
					motelistout+=arg0;					
				}				
			});
			try {
				cmd.start();
				cmd.waitFor();
			} catch (Exception e) {
				e.printStackTrace();
			}
			
			BufferedReader br=new BufferedReader(new StringReader(motelistout));
			String s,name,port,platform;
			Node node;
			try {
				while (((s=br.readLine())!=null)&&(s.indexOf(",")>=0)){
					StringTokenizer st=new StringTokenizer(s,",");
					if(!st.hasMoreTokens())break;
					name=st.nextToken();
					port=st.nextToken();
					platform="telosb";
					node=new Node(name,port,platform);
					result.addNode(node);
				}
			} catch (IOException e) {
				e.printStackTrace();
			}			
			
			return result;
		}
	}
	
	
	private static Command newMakeCommand(Node node,String dir,String makeopts,String makefile,String target,String platform,int deviceid){
		String opts[]={makeopts};
		return newMakeCommand(node, dir, opts, makefile, target, platform, deviceid);		
	}
	
	private static Command newMakeCommand(Node node,String dir,String makeopts[],String makefile,String target,String platform,int deviceid){
		Command cmd;
		File dirf=new File(dir);
		String[] args=new String[10];
		int argc=0;
		for(int i=0;i<makeopts.length;i++)
		if(!makeopts[i].equals("")){			
			args[argc++]=makeopts[i];
		}
		if(!makefile.equals("")){
			args[argc++]="-f";
			args[argc++]=makefile;
		}			
		args[argc++]=platform;
		
		if(!target.equals("")){
			args[argc++]=target;
			if(deviceid>=0){
				args[argc-1]+="."+deviceid;
			}
		}
		if(node!=null)
			args[argc++]="bsl,"+node.port;
		cmd=new RetriableCommand(new CommandLineCommand(
				"make",
				Arrays.copyOf(args,argc),
				dirf
		));
		if(target.equals("install")||target.equals("reinstall"))
			cmd=new SemaphoredCommand(serialSem,cmd);
		return cmd;		
	}	
	
	private static Command newIHex2BinCommand(String ihexfile){		
		RetriableCommand cmd;
		String args[]={ihexfile};
		cmd=new RetriableCommand(new CommandLineCommand(
				Environment.SCRIPTS_PATH+"/ihex_to_binary.sh",
				args,
				null
		));
		return cmd;		
	}	
	
	private static Command newGenerateKeysCommand(String file,int secParam,int keyUses,int encSecurityParam,int dosSecurityParam){		
		RetriableCommand cmd;
		String args[]={
				"net.tinyos.signet.SynapseKeyStorage",
				"-generate",
				Integer.toString(secParam),
				Integer.toString(keyUses),
				Integer.toString(encSecurityParam),
				Integer.toString(dosSecurityParam),
				file
		};
		cmd=new RetriableCommand(new CommandLineCommand(
				"java",
				args,
				null
		));
		return cmd;		
	}	
	private static Command newGetPublicKeysCommand(String file,String fileout){		
		RetriableCommand cmd;
		String args[]={
				"net.tinyos.signet.SynapseKeyStorage",
				"-get-public",
				file,
				fileout
		};
		cmd=new RetriableCommand(new CommandLineCommand(
				"java",
				args,
				null
		));
		return cmd;		
	}	
	
	private static Command newSendPublicKeysCommand(Node node,String file){		
		Command cmd;
		String args[]={
				"net.tinyos.signet.KeyVolumeManagerClient",
				"-comm",
				"serial@"+node.port+":"+node.platform,
				"-progress",
				"-upload",
				file
		};
		cmd=new RetriableCommand(new CommandLineCommand(
				"java",
				args,
				null
		));
		cmd=new SemaphoredCommand(serialSem,cmd);
		return cmd;		
	}		
	
	private static Command newCreateTagsCommand(String secretkeysfile,int blockSize,String unsignedfile,String signedfile,boolean doKeyUpdate,boolean doKeyRefresh,boolean lowOverheadKeyRefresh){		
		Command cmd;
		String args[]={
				"net.tinyos.signet.SecurityTaggerV0",
				"-sign",
				secretkeysfile,
				Integer.toString(blockSize),
				unsignedfile,
				signedfile,
				doKeyUpdate?"-update-keys":"",
				doKeyRefresh?"-use-key-refresh":"",
				lowOverheadKeyRefresh?"-low-overhead":""
		};
		cmd=new RetriableCommand(new CommandLineCommand(
				"java",
				args,
				null
		));
		return cmd;		
	}	
	
	private static Command newEncryptCommand(String secretkeysfile,String unsignedfile,String encryptedfile){		
		Command cmd;
		String args[]={
				"net.tinyos.signet.SecurityEncrypterV0",
				secretkeysfile,
				unsignedfile,
				encryptedfile
		};
		cmd=new RetriableCommand(new CommandLineCommand(
				"java",
				args,
				null
		));
		return cmd;		
	}		
	
	private static Command newFormatFlashCommand(Node node){		
		Command cmd;
		String args[]={
				"net.tinyos.signet.FlashManagerClient",
				"-comm",
				"serial@"+node.port+":"+node.platform,
				"-format"			
		};
		cmd=new RetriableCommand(new CommandLineCommand(
				"java",
				args,
				null
		));
		cmd=new SemaphoredCommand(serialSem,cmd);
		return cmd;		
	}	
	
	private static Command newStoreInFlashCommand(Node node,int id,String file,long tagsSize){		
		Command cmd;
		String args[]={
				"net.tinyos.signet.FlashManagerClient",
				"-comm",
				"serial@"+node.port+":"+node.platform,
				"-progress",
				"-print-table",
				"-writefile",
				Integer.toString(id,16),
				file,
				Long.toString(tagsSize,16)
		};
		cmd=new RetriableCommand(new CommandLineCommand(
				"java",
				args,
				null
		));
		cmd=new SemaphoredCommand(serialSem,cmd);
		return cmd;		
	}		
	
	private static class FileOutputPrinter implements OutputPrinter{
		FileWriter fstream;
		BufferedWriter out;
		public FileOutputPrinter(String file) {
			try {
				fstream = new FileWriter(file,true);
				out = new BufferedWriter(fstream);		
			} catch (IOException e) {
				e.printStackTrace();
			}	        	
		}
		
		@Override
		public void print(String arg0) {
			try {
				out.write(arg0);
				out.flush();
			} catch (IOException e) {
				e.printStackTrace();
			}		
		}
		
		@Override
		protected void finalize() throws Throwable {			
			out.close();
			super.finalize();
		}
	}
	
	
	private static Command newPrintfClientCommand(Node node,Semaphore endSem){		
		CommandLineCommand cmd;
		String args[]={
				"net.tinyos.tools.PrintfClient",
				"-comm",
				"serial@"+node.port+":"+node.platform,
		};
		cmd=new CommandLineCommand(
				"java",
				args,
				null
		);
		cmd.setTimeoutSemaphore(endSem,false);
		
		return cmd;		
	}		

	private static Command newSUINOCommand(Node basestation,String command,String param){		
		Command cmd;
		String args[]={
				"net.tinyos.signet.Suino",
				"-comm",
				"serial@"+basestation.port+":telosb",
				command,
				param
		};
		cmd=new RetriableCommand(new CommandLineCommand(
				//Environment.SUINO_SCRIPT_PATH,
				"java",
				args,
				null
		));
		cmd=new SemaphoredCommand(serialSem,cmd);
		return cmd;		
	}	
	
	private static Command newDelayCommand(long time){		
		return new DelayedCommand(time,new NullCommand());		
	}	
	
	
	private static class CommandExecutor implements Runnable{
		//private static int MAX_EXECUTORS=100;
		//private static Semaphore execsem=new Semaphore(MAX_EXECUTORS);
		
		private Semaphore endsem=null;
		private OutputPrinter op=null;
		private Queue<Command> cmdq=new LinkedList<Command>();
		
		public void postCommand(Command cmd){
			cmdq.add(cmd);
		}
		
		public void setEndingSemaphore(Semaphore sem){
			endsem=sem;
		}
		
		public void setOutputPrinter(OutputPrinter op){
			this.op=op;
		}
		
		public void setCommandsOutputPrinter(OutputPrinter op){
			for (Command cmd : cmdq) {
				if((op==null)||(cmd.op==null))
					cmd.setOutputPrinter(op);
			}
		}	
		
		private void printOut(String s){
			if(op!=null)
				op.print(s);
		}
		
		
		@Override
		public void run() {
			boolean errors=false;
			while(!cmdq.isEmpty()){
				try {
					//execsem.acquire();
					try{
					Command cmd=cmdq.remove();
					printOut(cmd.toString()+"\n");					
					boolean result=cmd.execute();
					printOut((result?"SUCCESS":"FAIL")+"\n");
					if(!result){
						errors=true;
						break;
					}
					}finally{
					//execsem.release();
					}
				} catch (Exception e) {
					e.printStackTrace();
					break;
				}				
			}
			cmdq.clear();
			printOut("Command sequence ended"+(errors?" WITH ERRORS":"")+".\n");
			if(endsem!=null)
				endsem.release();
		}
	}
	
	
	private static class Environment{
		public static final String HOME_PATH=System.getenv("HOME");
		public static final String SCRIPTS_PATH=System.getenv("MY_CONTRIB_DIR")+"/lib/signet/synapse/scripts";
		public static final String TOSROOT_PATH=System.getenv("TOSROOT");
		public static final String MY_CONTRIB_ROOT=System.getenv("MY_CONTRIB_ROOT");
		
		public static final String BLINK_APP="Blink";
		public static final String BLINK_APP_PATH=getTosAppPath("Blink");

		public static final String SYNAPSE_APP="Synapse";
		public static final String SYNAPSE_APP_PATH=getContribAppPath(SYNAPSE_APP);	
			
		//public static final String SUINO_SCRIPT="PingNode";
		public static final String SUINO_SCRIPT="Suino";
		public static final String SUINO_SCRIPT_PATH=SCRIPTS_PATH+"/"+SUINO_SCRIPT;
		public static final String SUINO_FORMAT="--format";
		public static final String SUINO_PREPARE="--prepare";
		public static final String SUINO_TRANSFER="--transfer";
		public static final String SUINO_LOAD="--load";
	
		public static final String SFM_APP="SynapseFlashManager";
		public static final String SFM_APP_PATH=getContribAppPath(SFM_APP);		
		
		public static final String FM_APP="FlashManager";
		//public static final String FM_APP=SFM_APP;
		public static final String FM_APP_PATH=getContribAppPath(FM_APP);		
		
		public static final String KVM_APP="KeyVolumeManager";
		//public static final String KVM_APP=SFM_APP;
		public static final String KVM_APP_PATH=getContribAppPath(KVM_APP);
	
		
		// TODO: Decidere locazione di questi file di configurazione
		public static final String CONFIG_FILE=HOME_PATH+"/synapse-config.txt";
		public static final String SECRET_KEYS_FILE=HOME_PATH+"/synapse-secret-keys.xml";
		public static final String PUBLIC_KEYS_FILE=HOME_PATH+"/synapse-public-keys.bin";
		
		public static String getAppBuildPath(String appPath,String platform){
			return appPath+"/build/"+platform;
		}
		public static String getAppIHexPath(String appPath,String platform){
			return getAppBuildPath(appPath,platform)+"/main.ihex";
		}
		public static String getAppBinaryPath(String appPath,String platform){
			return getAppBuildPath(appPath,platform)+"/main.ihex.compact.binary";
		}
		public static String getAppEncryptedBinaryPath(String appPath,String platform){
			return getAppBuildPath(appPath,platform)+"/main.encrypted.binary";
		}		
		public static String getAppSignedBinaryPath(String appPath,String platform){
			return getAppBuildPath(appPath,platform)+"/main.signed.binary";
		}		
		
		public static String getTosAppPath(String appname){
			return TOSROOT_PATH+"/apps/"+appname;
		}
		public static String getTosAppBuildPath(String appname,String platform){
			return getAppBuildPath(getTosAppPath(appname),platform);
		}
		public static String getTosAppIHexPath(String appname,String platform){
			return getAppIHexPath(getTosAppPath(appname),platform);
		}		
		public static String getTosAppBinaryPath(String appname,String platform){
			return getAppBinaryPath(getTosAppPath(appname),platform);
		}
		public static String getTosAppSignedBinaryPath(String appname,String platform){
			return getAppSignedBinaryPath(getTosAppPath(appname),platform);
		}			
		
		public static String getContribAppPath(String appname){
			return MY_CONTRIB_ROOT+"/apps/"+appname;
		}		
	}
	
	private static class NumberedOutputPrinter implements OutputPrinter{
		private OutputPrinter op;
		private int num;
		
		public NumberedOutputPrinter(int num,OutputPrinter op){
			this.op=op;
			this.num=num;
		}
		
		@Override
		public void print(String arg0) {
			if(op!=null)
				op.print(""+num+" : "+arg0);
		}
	}
	
	private static void executeParallel(CommandExecutor[] executors){
		// Wait for all nodes
		Semaphore endsem=new Semaphore(-executors.length+1);
		for (CommandExecutor cmdex : executors) {
			cmdex.setEndingSemaphore(endsem);
			new Thread(cmdex).start();			
		}
		try {
			endsem.acquire();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}		
	}
	
	private static CommandExecutor createExecutor(Command[] commands, OutputPrinter op){
		CommandExecutor cmdex=new CommandExecutor();
		for(Command cmd : commands){
			cmdex.postCommand(cmd);
		}
		cmdex.setCommandsOutputPrinter(op);
		return cmdex;
	}
	
	private static class Configuration{
		
		private static Configuration mainInstance;
		
		public static Configuration getInstance(){
			if(mainInstance==null){
				mainInstance=new Configuration();
			}
			return mainInstance;
		}
		
		public int securityBits;
		public int securityBitsLog(){
			return (int)Math.ceil(Math.log(securityBits+1)/Math.log(2));
		}
		
		public void read(){
			FileReader fstream;
			BufferedReader in;
			try {
				fstream = new FileReader(Environment.CONFIG_FILE);
				in = new BufferedReader(fstream);
				securityBits=Integer.parseInt(in.readLine());				
				in.close();
			} catch (IOException e) {
				e.printStackTrace();
			}			
		}
		public void write(){
			FileWriter fstream;
			BufferedWriter out;
			try {
				fstream = new FileWriter(Environment.CONFIG_FILE,true);
				out = new BufferedWriter(fstream);
				out.write(Integer.toString(securityBits)+"\n");				
				out.flush();	
				out.close();
			} catch (IOException e) {
				e.printStackTrace();
			}				
		}	
	}
	
	private static boolean useSecurity(){
		return useAuthentication || useEncryption || useDoSProtection;
	}
	
	public static void deployment(int secParam,int keyUses,int encSecParam,int dosSecParam,boolean compileTools,boolean keepKeys){
		Semaphore endsem;
		Network net=new NodeDiscovery().discover();
		checkNodes(net);
		CommandExecutor cmdex=new CommandExecutor();
		OutputPrinter op=sysop;
		
		Configuration.getInstance().securityBits=secParam;
		Configuration.getInstance().write();
		
		if(useSecurity()){
			// Generate keys
			if(!keepKeys)
				cmdex.postCommand(newGenerateKeysCommand(Environment.SECRET_KEYS_FILE,secParam,keyUses,encSecParam,dosSecParam));
			// Retrieve public keys
			cmdex.postCommand(newGetPublicKeysCommand(Environment.SECRET_KEYS_FILE,Environment.PUBLIC_KEYS_FILE));
			// Make KeyVolumeManager application
			// TODO: multiple platforms concurrently
			if(compileTools)
				cmdex.postCommand(newMakeCommand(null,Environment.KVM_APP_PATH,"","","","telosb",-1));
				//cmdex.postCommand(newMakeCommand(null,Environment.SFM_APP_PATH,"","","","telosb",-1));
		}
		// Make Synapse application - always because it could be compiled with adversary flag
		String opts[]={
			"SECURITY_BITS="+Configuration.getInstance().securityBits,
			"SECURITY_BITS_LOG="+Configuration.getInstance().securityBitsLog(),
			"IS_NODE=TRUE",
			useDoSProtection?"USE_DOS_PROTECTION=TRUE":"",
			useEncryption?"USE_CI_ENCRYPTION=TRUE":"",
			useAuthentication?"USE_AUTHENTICATION=TRUE":""
		};
		
		cmdex.postCommand(newMakeCommand(
			null,Environment.SYNAPSE_APP_PATH,opts,
			"","","telosb",-1));		
//		cmdex.postCommand(newIHex2BinCommand(Environment.getAppIHexPath(Environment.SYNAPSE_APP_PATH, "telosb")));
		
		cmdex.setOutputPrinter(op);
		cmdex.setCommandsOutputPrinter(op);
		
		cmdex.run();
		
		// Wait for all nodes
		endsem=new Semaphore(-net.nodes.size() +1);
		
		// Deploy keys and synapse in each node
		int id=1;
		for (Node node : net.getNodes()) {
			OutputPrinter nop;
			cmdex=new CommandExecutor();		
			if(node!=net.getBaseStation()){
				id++;
				nop=new NumberedOutputPrinter(id,op);									
			}else{
				nop=new NumberedOutputPrinter(1,op);
			}
			
			if(useSecurity()){
				cmdex.postCommand(newMakeCommand(node,Environment.KVM_APP_PATH,"","","reinstall","telosb",id));
				cmdex.postCommand(newSendPublicKeysCommand(node,Environment.PUBLIC_KEYS_FILE));
			}
			cmdex.postCommand(newMakeCommand(
					node,Environment.SYNAPSE_APP_PATH,"",
					"","reinstall","telosb",
					node==net.getBaseStation()?1:id
					));
			cmdex.setOutputPrinter(nop);
			cmdex.setCommandsOutputPrinter(nop);
			cmdex.setEndingSemaphore(endsem);
			new Thread(cmdex).start();			
		}
		try {
			endsem.acquire();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		System.out.print("All nodes done.\n");
		
	}
	
	public static void deployAdversaries(String bogusAppPath,int bogusAppId,boolean compileTools){
		Semaphore endsem;
		Network net=new NodeDiscovery().discover();
		checkNodes(net);
		CommandExecutor cmdex=new CommandExecutor();
		OutputPrinter op=sysop;
		
		Configuration.getInstance().read();
		
		// Bogus application is supposed to be already built		
//		cmdex.postCommand(newMakeCommand(
//				null,
//				bogusAppPath,"","","","telosb",-1));
		
		// Make FlashManager application
		if(compileTools)		
			cmdex.postCommand(newMakeCommand(
				null,
				Environment.FM_APP_PATH,"","","","telosb",-1));		
		// Make Synapse application - always because it could be compiled with adversary flag
		String opts[]={
				"SECURITY_BITS="+Configuration.getInstance().securityBits,
				"SECURITY_BITS_LOG="+Configuration.getInstance().securityBitsLog(),
				"IS_NODE=TRUE",
				"ADVERSARY=TRUE",
			};		
		cmdex.postCommand(newMakeCommand(
			null,Environment.SYNAPSE_APP_PATH,
			opts,
			"","","telosb",-1));		
		
		cmdex.setOutputPrinter(op);
		cmdex.setCommandsOutputPrinter(op);
		
		cmdex.run();
		
		// Wait for all nodes
		endsem=new Semaphore(-net.nodes.size() +1);
		
		int id=1;
		for (Node node : net.getNodes()) {			
			id++;
			cmdex=new CommandExecutor();
			OutputPrinter nop=new NumberedOutputPrinter(1,op);
			// Deploy FlashManager
			cmdex.postCommand(newMakeCommand(
					node,
					Environment.FM_APP_PATH,"","","reinstall","telosb",-1));
			// Format flash
			cmdex.postCommand(newFormatFlashCommand(node));
			// Deploy bogus app
			cmdex.postCommand(newStoreInFlashCommand(
					node, bogusAppId, 
					Environment.getAppBinaryPath(bogusAppPath, "telosb"),0));			
			// Deploy Synapse
			cmdex.postCommand(newMakeCommand(
					node,Environment.SYNAPSE_APP_PATH,"",				
					"","reinstall","telosb",
					id
					));
			
			cmdex.setOutputPrinter(nop);
			cmdex.setCommandsOutputPrinter(nop);
			cmdex.setEndingSemaphore(endsem);
			new Thread(cmdex).start();			
		}
		try {
			endsem.acquire();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		System.out.print("All nodes done.\n");
		
	}	
	
	public static void deploySynapse(boolean compileTools){
		Semaphore endsem;
		Network net=new NodeDiscovery().discover();
		checkNodes(net);
		CommandExecutor cmdex=new CommandExecutor();
		OutputPrinter op=sysop;
		Node bs=net.getBaseStation();
		
		Configuration.getInstance().read();
		
		// Make Synapse application for nodes
		String opts[]={
				"SECURITY_BITS="+Configuration.getInstance().securityBits,
				"SECURITY_BITS_LOG="+Configuration.getInstance().securityBitsLog(),
				"IS_NODE=TRUE",
				useDoSProtection?"USE_DOS_PROTECTION=TRUE":"",
				useEncryption?"USE_CI_ENCRYPTION=TRUE":"",
				useAuthentication?"USE_AUTHENTICATION=TRUE":""				
			};			
		cmdex.postCommand(newMakeCommand(
			null,Environment.SYNAPSE_APP_PATH,
			opts,
			"","","telosb",-1));		
				
		cmdex.setOutputPrinter(op);
		cmdex.setCommandsOutputPrinter(op);		
		
		cmdex.run();
		
		// Wait for all nodes except base station
		endsem=new Semaphore(-(net.nodes.size()-1) +1);
		
		// Deploy synapse in each node
		int id=1;
		for (Node node : net.getNodes()) {
			if(node!=bs){
				OutputPrinter nop;
				cmdex=new CommandExecutor();		
				if(node!=net.getBaseStation()){
					id++;
					nop=new NumberedOutputPrinter(id,op);				
				}else{
					nop=new NumberedOutputPrinter(1,op);
				}
				cmdex.postCommand(newMakeCommand(
						node,Environment.SYNAPSE_APP_PATH,"","","reinstall","telosb",
						node==net.getBaseStation()?1:id
						));
				cmdex.setOutputPrinter(nop);
				cmdex.setCommandsOutputPrinter(nop);
				cmdex.setEndingSemaphore(endsem);
				new Thread(cmdex).start();
			}
		}
		try {
			endsem.acquire();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		System.out.print("All nodes done.\n");		
	}	
	
	public static void deploynode(String port,int id, boolean compileTools){
		CommandExecutor cmdex=new CommandExecutor();
		OutputPrinter op=sysop;
		Network net=new NodeDiscovery().discover();
		Node node=net.findNodeByPort(port);
		if(node==null){
			System.out.print("Node not found\n.");
			return;
		}
		
		Configuration.getInstance().read();
		
		// Retrieve public keys
		cmdex.postCommand(newGetPublicKeysCommand(Environment.SECRET_KEYS_FILE,Environment.PUBLIC_KEYS_FILE));
		// Install KeyVolumeManager application
		// TODO: multiple platforms concurrently
		if(compileTools)		
			cmdex.postCommand(newMakeCommand(null,Environment.KVM_APP_PATH,"","","","telosb",-1));
		cmdex.postCommand(newMakeCommand(node,Environment.KVM_APP_PATH,"","","reinstall","telosb",-1));
		// Send public keys
		cmdex.postCommand(newSendPublicKeysCommand(node,Environment.PUBLIC_KEYS_FILE));
		// Make Synapse application - always because it could be compiled with adversary flag
		String opts[]={
				"SECURITY_BITS="+Configuration.getInstance().securityBits,
				"SECURITY_BITS_LOG="+Configuration.getInstance().securityBitsLog(),
				"IS_NODE=TRUE",
				useDoSProtection?"USE_DOS_PROTECTION=TRUE":"",
				useEncryption?"USE_CI_ENCRYPTION=TRUE":"",
				useAuthentication?"USE_AUTHENTICATION=TRUE":""	
			};	
		cmdex.postCommand(newMakeCommand(
				null,Environment.SYNAPSE_APP_PATH,
				opts,
				"","","telosb",-1));
		cmdex.postCommand(newMakeCommand(node,Environment.SYNAPSE_APP_PATH,"","","reinstall","telosb",id));
				
		cmdex.setOutputPrinter(op);
		cmdex.setCommandsOutputPrinter(op);
		
		cmdex.run();
		
		System.out.print("Done.\n");		
	}
	
	public static void dissemination(String appPath,int blockSize,boolean updateKeys,boolean keyRefresh,boolean lowOverheadKeyRefresh,boolean formatBaseStation,int appid,boolean formatNodes,boolean deploySynapse,boolean compileTools,boolean listenPrintf, String printfdir, boolean interactiveEnd){
		Semaphore endsem;		
		String platform="telosb";//TODO: Implement multi-platform
		Network net=new NodeDiscovery().discover();
		checkNodes(net);

		OutputPrinter op=sysop;
		Node bs=net.getBaseStation();
	
		Configuration.getInstance().read();
		
		CommandExecutor executors[]={
			new CommandExecutor(),
			new CommandExecutor()
		};
			
		executors[0].postCommand(newIHex2BinCommand(
				Environment.getAppIHexPath(appPath,platform)));		
		
		
		String applicationFile=Environment.getAppBinaryPath(appPath,platform);
		String fileToDisseminate=applicationFile;	
		
		if(useEncryption){
			String encFile=Environment.getAppEncryptedBinaryPath(appPath,platform);
			executors[0].postCommand(newEncryptCommand(
					Environment.SECRET_KEYS_FILE,
					fileToDisseminate,
					encFile
			));				
			fileToDisseminate=encFile;
		}
		if(useAuthentication){
			String signedFile=Environment.getAppSignedBinaryPath(appPath,platform);
			executors[0].postCommand(newCreateTagsCommand(
					Environment.SECRET_KEYS_FILE,
					blockSize,
					fileToDisseminate,
					signedFile,
					updateKeys,keyRefresh,lowOverheadKeyRefresh));
			fileToDisseminate=signedFile;
		}
				
		if(compileTools)
			executors[1].postCommand(newMakeCommand(
				null,
				Environment.FM_APP_PATH,"","","","telosb",-1));
		executors[1].postCommand(newMakeCommand(
				bs,
				Environment.FM_APP_PATH,"","","reinstall","telosb",-1));		
		if(formatBaseStation)
			executors[1].postCommand(newFormatFlashCommand(bs));
				
		executors[0].setCommandsOutputPrinter(new NumberedOutputPrinter(1,op));
		executors[1].setCommandsOutputPrinter(new NumberedOutputPrinter(2,op));		
		executeParallel(executors);
		
		//TODO:Include this functionality in a command to speed up.
		long appSize=new File(fileToDisseminate).length();
		long tagsSize=
			appSize
			-
			new File(applicationFile).length()
		;	
		
		executors[0].postCommand(newStoreInFlashCommand(
				bs, appid, 
				fileToDisseminate,
				tagsSize));		

		// Make Synapse application for base station
		{
		String opts[]={
				"SECURITY_BITS="+Configuration.getInstance().securityBits,
				"SECURITY_BITS_LOG="+Configuration.getInstance().securityBitsLog(),
				"IS_BASESTATION=TRUE",
				useDoSProtection?"USE_DOS_PROTECTION=TRUE":"",
				useEncryption?"USE_CI_ENCRYPTION=TRUE":"",
				useAuthentication?"USE_AUTHENTICATION=TRUE":""		
			};
		executors[1].postCommand(newMakeCommand(
			null,Environment.SYNAPSE_APP_PATH,
			opts,			
			"","","telosb",-1));		
		}
		executors[0].setCommandsOutputPrinter(new NumberedOutputPrinter(1,op));
		executors[1].setCommandsOutputPrinter(new NumberedOutputPrinter(2,op));		
		executeParallel(executors);
		
		CommandExecutor cmdex=executors[0];		
		cmdex.setEndingSemaphore(null);
		// Deploy synapse in base station
		cmdex.postCommand(newMakeCommand(
				bs,Environment.SYNAPSE_APP_PATH,"",
				"","reinstall","telosb",65534));		
		
		cmdex.setOutputPrinter(op);
		cmdex.setCommandsOutputPrinter(op);
		cmdex.run();
		
		
		
		final int BOOTSTRAP_DELAY=15000;
		final int PREPARE_DELAY=10000;
		final int FORMAT_DELAY=80000;
		final int DISSEMINATION_CYCLE_DELAY=1500;
		
		int DISSEMINATION_DELAY=(int)(
			(appSize/blockSize)*2*DISSEMINATION_CYCLE_DELAY
		);		
		
		if(deploySynapse){	
			
			// Make Synapse application for nodes
			cmdex=executors[0];
			cmdex.setEndingSemaphore(null);			
			{
			String opts[]={
					"SECURITY_BITS="+Configuration.getInstance().securityBits,
					"SECURITY_BITS_LOG="+Configuration.getInstance().securityBitsLog(),
					"IS_NODE=TRUE",
					useDoSProtection?"USE_DOS_PROTECTION=TRUE":"",
					useEncryption?"USE_CI_ENCRYPTION=TRUE":"",
					useAuthentication?"USE_AUTHENTICATION=TRUE":""					
				};
			cmdex.postCommand(newMakeCommand(
					null,Environment.SYNAPSE_APP_PATH,
					opts,			
					"","","telosb",-1));							
			}							
			cmdex.setOutputPrinter(op);
			cmdex.setCommandsOutputPrinter(op);
			cmdex.run();
			
			// Wait for all nodes except bs
			endsem=new Semaphore(-(net.nodes.size()-1) +1);		
		
			// Deploy synapse in each node
			int nodeid=1;
			for (Node node : net.getNodes()) {			
				if(node!=bs){
					nodeid++;
					CommandExecutor loopcmdex=new CommandExecutor();
				
					OutputPrinter nop=new NumberedOutputPrinter(nodeid,op);						
					
					loopcmdex.postCommand(newMakeCommand(
							node,Environment.SYNAPSE_APP_PATH,"","","reinstall","telosb",
							nodeid
							));				
					
					loopcmdex.setOutputPrinter(nop);
					loopcmdex.setCommandsOutputPrinter(nop);
					loopcmdex.setEndingSemaphore(endsem);
					new Thread(loopcmdex).start();
				}
			}		
			
			try {
				endsem.acquire();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			System.out.print("All nodes done.\n");
		}	
		

		
		cmdex.postCommand(newDelayCommand(BOOTSTRAP_DELAY));
		if(formatNodes){		
			cmdex.postCommand(newSUINOCommand(
					bs,Environment.SUINO_PREPARE,""));			
			
			cmdex.postCommand(new DelayedCommand(PREPARE_DELAY,newSUINOCommand(
					bs,Environment.SUINO_FORMAT,"")));
				
			cmdex.postCommand(newDelayCommand(FORMAT_DELAY));
		}
			
		Semaphore disseminationEnd=new Semaphore(0);
		
		if(listenPrintf){		
			
			int nodeid=1;
			for (Node node : net.getNodes()) {			
				if(node!=bs){
					nodeid++;
					CommandExecutor loopcmdex=new CommandExecutor();
				
					OutputPrinter fop=new FileOutputPrinter(printfdir+"/printfout-"+nodeid+".txt");						
					
					loopcmdex.postCommand(newPrintfClientCommand(
							node, disseminationEnd));				
					
					loopcmdex.setOutputPrinter(fop);
					loopcmdex.setCommandsOutputPrinter(fop);

					new Thread(loopcmdex).start();
				}
			}	
			try {
				Thread.sleep(1000);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}			
		}	
		
		cmdex.postCommand(newSUINOCommand(bs,Environment.SUINO_TRANSFER,Integer.toString(appid,16)));
		
		
/*		
		if(interactiveEnd)
			cmdex.postCommand(new InteractiveWaitCommand("Wait dissemination to complete."));
		else
			cmdex.postCommand(newDelayCommand(DISSEMINATION_DELAY));
*/		
		cmdex.setOutputPrinter(op);
		cmdex.setCommandsOutputPrinter(op);		
		cmdex.run();
		
		
		disseminationEnd.release();
		
		cmdex.postCommand(newSUINOCommand(
				bs,Environment.SUINO_PREPARE,""));
		
		cmdex.postCommand(new DelayedCommand(PREPARE_DELAY,newSUINOCommand(
				bs,Environment.SUINO_LOAD,Integer.toString(appid,16))));

		cmdex.setOutputPrinter(op);
		cmdex.setCommandsOutputPrinter(op);		
		cmdex.run();
		
		
	}
	
	private static void compileTools(){
		String platform="telosb";//TODO: Implement multi-platform
		CommandExecutor cmdex=new CommandExecutor();
			
		CommandExecutor exs[]=new CommandExecutor[2];
		Command cmds[]=new Command[1];
		cmds[0]=newMakeCommand(
				null,
				Environment.FM_APP_PATH,"","","",platform,-1);
		exs[0]=createExecutor(cmds,new NumberedOutputPrinter(1,sysop));
		
		cmds=new Command[1];
		cmds[0]=newMakeCommand(
				null,
				Environment.KVM_APP_PATH,"","","",platform,-1);
		exs[1]=createExecutor(cmds,new NumberedOutputPrinter(2,sysop));		
		
		executeParallel(exs);		
	}
	
	private static void checkNodes(Network net){
		if(net.nodes.size()==0){
			System.out.println("No nodes detected with motelist.");
			System.exit(1);
		}			
	}
	
	public static void printUsage(){		
		System.out.println(
			"Usage:\n"+
			"java "+SecureSynapseInterface.class.getName()+"\n"+
			"\t[-deploy <auth. security param.> <T> <encryption security param> <DoS protection security param> [-keep-keys]]\n"+
//			"\t[-deploy-node <node port> <node id>]\n"+
//			"\t[-deploy-adversaries <bogus application path> <bogus application id>]\n"+
//			"\t[-dissem <application path> <block size> <application id> [-use-key-refresh [-low-overhead]] [-format-nodes] [-format-bs] [-listen-printf <output directory>] [-interactive-end]]\n"+
			"\t[-dissem <application path> <block size> <application id> [-use-key-refresh [-low-overhead]] [-format-nodes] [-format-bs] ]\n"+
//			"\t[-deploy-synapse]\n"+
			"\t[-compile-tools]\n"+
			"\t[-use-authentication]\n"+			
			"\t[-use-encryption]\n"+
			"\t[-use-dos-protection]\n"+
			"See javadoc for detailed information.\n"
		);
	}
	
	/**
	 * Command line interface to this class.
	 * Commands accepted:
	 * <ul>
	 * <li>
	 * <code>-deploy &lt;security param.> &lt;T> </code>
	 * <br>
	 * With this command you can deploy everything needed for secure code update with Synapse.<br>
	 * Application automatically discovers nodes with motelist and considers base station the 
	 * node with lowest id.<br>  
	 * <ul>
	 * <li>Generates a new couple of secret+public keys with given security parameter and T.</li>
	 * <li>Stores private keys in home directory.</li>
	 * <li>Deploys public keys in nodes using KeyVolumeManager.</li>
	 * <li>Installs Synapse on nodes</li>
	 * </ul>
	 * </li>
	 * <li>
	 * <code>-deploy-node &lt;node port> <node id> </code><br>
	 * With this command you can add a node to the network anytime after the network has been deployed.
	 * <ul>
	 * <li>Uses private keys stored in home directory to get current public keys.</li>
	 * <li>Deploys public keys node on given port using KeyVolumeManager.</li>
	 * <li>Installs Synapse on nodes assigning the given id.</li>
	 * </ul>
	 * </li>
	 * <li>
	 * <code>-deploy-adversaries &lt;bogus application path> &lt;bogus application id> </code><br>
	 * With this command you can deploy on a node what needed to simulate an adversary node
	 * that wants to corrupt the current dissemination session.<br>
	 * You need to specify the bogus application path and the desired application id.<br>
	 * The application is supposed to be already built.<br>
	 * All nodes detected via motelist will be programmed as adversaries.
	 * </li>
	 * <li>
	 * <code>-deploy-synapse</code><br>
	 * With this command you can install Synapse in all nodes connected.
	 * </li>
	 * <li>
	 * <code>-dissem &lt;application path> &lt;block size> &lt;application id> [-use-key-refresh [-low-overhead]] [-format-nodes] [-format-bs] [-listen-printf &lt;output directory>]</code><br>
	 * With this command you can start an authenticated dissemination session.
	 * You need to specify:
	 * <ul>
	 * <li>the path to the application to be disseminated,</li>
	 * <li>the size of the blocks used in Synapse transfers,</li>
	 * <li>the desired application id,</li>
	 * <li>the path to the application to be disseminated,</li>
	 * <li>With <code>-use-key-refresh</code> option set, partial key refreshment for 
	 * protection against strong adversary is used.<br>
	 * If you set also <code>-low-overhead</code> option, a low overhead key refreshment
	 * system is used, but this will not protect against strong adversary.  
	 * </li>
	 * <li>With <code>-format-nodes</code> option set, nodes will be given the format command 
	 * before the transfer.   
	 * </li>
	 * <li>With <code>-format-bs</code> option set the base station will be formatted before transferring 
	 * the application on it.
	 * </li>
	 * <li>With <code>-listen-printf</code> option set, output from nodes will be saved in files under 
	 * given directory. 
	 * </li>
	 * <li>With <code>-timeout-end</code> option set, dissemination is stopped after an amount of time calculated from app. size, 
	 * assuming single hop dissemination, and absence of adversaries.<br>
	 * Otherwise, user will be asked to press a key when dissemination finishes.<br>
	 * </li>
	 * </ul>
	 * The application is supposed to be already built.<br>  
	 * </li>
	 * <li>
	 * <code>-compile-tools</code><br>
	 * If only this option is set, all tinyos apps needed will be built.<br>
	 * This must be done if tools has never been built.<br> 
	 * This option can also be specified together with all the commands.<br>
	 * If set together with any command, all tinyos needed for the command will be 
	 * recompiled before installation.
	 * This should be useful only for developers of the tools. 
	 * </li> 
	 * </ul>
	 * General options accepted: 
	 * <ul>
	 * <li>
	 * <code>-use-encryption</code><br>
	 * If this option is set, then encryption will be used to ensure confidentiality of the 
	 * disseminated code image 
	 * </li>  
	 * <li>
	 * <code>-use-dos-protection</code><br>
	 * If this option is set, then dissemination will be protected against DoS attacks.
	 * </li>
	 * <li>
	 * <code>-dont-use-security</code><br>
	 * If this option is set, then security (signature,encryption,dos protection) will be disabled
	 * </li>  
	 * </ul>
	 */
	public static void main(String[] args) throws Exception{
		boolean deploy=false,deploynode=false,deployadv=false,dissem=false;
		boolean keyrefresh=false,useLowOverheadKeyRefresh=false,formatbs=false,formatnodes=false,deploySynapse=false,compileTools=false,listenPrintf=false,interactiveEnd=true,keepKeys=false;
		int secParam=80,useCount=30,blockSize=800,appid=0xAAA,nodeid=0xFF,bogusAppId=0xAAAA,encSecParam=128,dosSecParam=128;
		String appPath="",port="",bogusAppPath="",printfOut="";
		int argi=0;		
		if(args.length==0){
			printUsage();
			System.exit(1);			
		}
		while(argi<args.length){
			if      (args[argi].equals("-deploy")){
				deploy=true;
				secParam=Integer.parseInt(args[++argi]);
				useCount=Integer.parseInt(args[++argi]);
				encSecParam=Integer.parseInt(args[++argi]);
				dosSecParam=Integer.parseInt(args[++argi]);
			}else if(args[argi].equals("-deploy-node")){
				deploynode=true;
				port=args[++argi];
				nodeid=Integer.parseInt(args[++argi]);
			}else if(args[argi].equals("-deploy-adversaries")){
				deployadv=true;
				bogusAppPath=args[++argi];
				bogusAppId=Integer.parseInt(args[++argi],16);				
			}else if(args[argi].equals("-dissem")){
				dissem=true;
				appPath=args[++argi];
				blockSize=Integer.parseInt(args[++argi]);
				appid=Integer.parseInt(args[++argi],16);
			}else if(args[argi].equals("-listen-printf")){
				listenPrintf=true;
				printfOut=args[++argi];
			}else if(args[argi].equals("-use-key-refresh")){
				keyrefresh=true;
			}else if(args[argi].equals("-low-overhead")){	
				useLowOverheadKeyRefresh=true; 
			}else if(args[argi].equals("-format-bs")){
				formatbs=true;
			}else if(args[argi].equals("-format-nodes")){
				formatnodes=true;
			}else if(args[argi].equals("-deploy-synapse")){
				deploySynapse=true;
			}else if(args[argi].equals("-timeout-end")){
				interactiveEnd=false;
			}else if(args[argi].equals("-use-encryption")){
				useEncryption=true;
			}else if(args[argi].equals("-use-dos-protection")){
				useDoSProtection=true;
			}else if(args[argi].equals("-use-authentication")){
				useAuthentication=true;
			}else if(args[argi].equals("-keep-keys")){
				keepKeys=true;
			}else if(args[argi].equals("-compile-tools")){
				compileTools=true;				
			}else{
				printUsage();
				System.exit(1);
			}
			argi++;
		}		
		
/*		
		if(!useSecurity){
			if(useDoSProtection || useEncryption){
				System.err.print("Since security is disabled, neither encryption nor dos protection are enabled.");
			}
			useDoSProtection=false;
			useEncryption=false;
		}
*/		
		if(deploy)		
			deployment(secParam,useCount,encSecParam,dosSecParam,compileTools,keepKeys);
		else if(!dissem && deploySynapse)
			deploySynapse(compileTools)
		;
		if(deployadv){
			deployAdversaries(bogusAppPath, bogusAppId, compileTools);
		}
		if(deploynode)
			deploynode(port,nodeid, compileTools);
		if(dissem)
			dissemination(appPath,blockSize,true,keyrefresh,useLowOverheadKeyRefresh,formatbs,appid,formatnodes,deploySynapse, compileTools, listenPrintf,printfOut,interactiveEnd);
		
		if(compileTools && (args.length==1))
			compileTools();
			
	}

}
