/* Copyright (c) 2009 The Regents of the Signet Group.
*  All rights reserved.
*
*  Permission to use, copy, modify, and distribute this software and its
*  documentation for any purpose, without fee, and without written
*  agreement is hereby granted, provided that the above copyright
*  notice, the (updated) modification history and the author appear in
*  all copies of this source code.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS `AS IS'
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
*  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS
*  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, LOSS OF USE, DATA,
*  OR PROFITS) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
*  THE POSSIBILITY OF SUCH DAMAGE.
*/

package net.tinyos.signet;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.security.InvalidParameterException;
import java.util.Arrays;

import javax.xml.parsers.ParserConfigurationException;

import org.bouncycastle.crypto.engines.AESEngine;
import org.xml.sax.SAXException;

import sun.nio.ch.DevPollSelectorProvider;

import com.sun.crypto.provider.BlowfishCipher;

/**
 * Adds security tags to a given file.
 * Resulting file format is described in security_tags.h, in TinyOS component SecurityVerifierV0
 * Signature is variable-length, and the message to sign is variable-length too, because it's 
 * the remaining space of a fixed length block containing signature.
 * So i must guess signature size to calculate the message to be signed, and then if my guess 
 * was right i can use the signature, otherwise i have to guess another signature length and retry. 
 * To increase success probability, and to allow to search for short signatures, i introduce a 
 * variable byte in the message, so i can try with 256 values of this byte to check my guess. 
 * @author Moreno Dissegna
 */
public class SecurityTaggerV0 {
	
	// Dimensione della struttura SessionInfo definita in security_tags.h
	private static int ST0_SESSION_INFO_SIZE=7;
	
	//private static HashFunction hf=new SHA1HashFunction();
	private static HashFunction hf=new DaviesMeyerHashConstruction(new AESEngine());

	// Signature algorithm
	private TTimeSA sa;
	// Secret keys
	private TTimeSA.SecretKeyRing skr;
	// Security parameter protecting message authentication
	private int securityParam;
	// Whether to use the key refreshment system.
	private boolean useKeyRefresh;
	// Whether to use the low overhead key refreshment system.
	private boolean useLowOverheadKeyRefresh;	
	
	
	public void setUseKeyRefresh(boolean useKeyRefresh) {
		this.useKeyRefresh = useKeyRefresh;
	}
	public boolean getUseKeyRefresh() {
		return useKeyRefresh;
	}
	/**
	 * @param useLowOverheadKeyRefresh the useLowOverheadKeyRefresh to set
	 */
	public void setUseLowOverheadKeyRefresh(boolean useLowOverheadKeyRefresh) {
		this.useLowOverheadKeyRefresh = useLowOverheadKeyRefresh;
	}
	/**
	 * @return the useLowOverheadKeyRefresh
	 */
	public boolean isUseLowOverheadKeyRefresh() {
		return useLowOverheadKeyRefresh;
	}
	/**
	 * Creates new object with a newly created keyring, with given securityParam and keyUses,
	 * that will be stored in the given file.
	 * @param privateKeysFileName
	 * @param securityParam
	 * @param keyUses
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException 
	 */
	public SecurityTaggerV0(String privateKeysFileName, int securityParam, int keyUses) throws IOException, ParserConfigurationException, SAXException {
		this.securityParam=securityParam;
		sa=new TTimeSA(hf);
		skr=sa.generateNewKeyRing(getKeySize(), keyUses);
		skr.saveToFile(privateKeysFileName);		
	}
	/**
	 * Creates new object using the keyring stored in the given file
	 * @param privateKeysFileName
	 * @throws SAXException
	 * @throws IOException
	 * @throws ParserConfigurationException
	 */
	public SecurityTaggerV0(String privateKeysFileName) throws SAXException, IOException, ParserConfigurationException {
		sa=new TTimeSA(hf);
		skr=sa.loadKeyRing(privateKeysFileName);
		this.securityParam=skr.keyLength*8;		
	}	
	
	/**
	 * @return Size of hashes
	 */
	private int getHashSize(){
		return securityParam/8;
	}
	/**
	 * @return Size of keys
	 */
	private int getKeySize(){
		return securityParam/8;
	}
	/**
	 * @return Maximum length of the signature, measured in number of hashes it contains
	 */
	private int getSigMaxSize(){
		return securityParam+ (int)(Math.log(securityParam)/Math.log(2)) +1;
	}
	/**
	 * Start value when i try to guess signature size.
	 * On average signature length should be about half the maximum size, but i look for
	 * short signatures, so i start from guessing short lengths.
	 * @return Start value when i try to guess signature size.
	 */
	private int getStartGuessSigHashes(){
		return getSigMaxSize()/4;
	}	
	/**
	 * Size of the header containing the signature
	 * @param sigKeys Number of keys in the signature
	 * @return Size of the header containing the signature
	 */
	private int getSignatureHeaderSize(int sigKeys){
		return 
			1+
			sigKeys*getKeySize()
			;
	}
	/**
	 * Size of header containing signature and session info, depending on signature length
	 * @param sigLength
	 * @return
	 */
	private int getHeaderSize(int sigKeys){
		return 
			getSignatureHeaderSize(sigKeys)+
			ST0_SESSION_INFO_SIZE
		;
	}
	/**
	 * @param blocksCount
	 * @return Size of hashes of blocks section
	 */
	private int getHashesSize(int blocksCount){
		return blocksCount*getHashSize();
	}
	
	
	/**
	 * @return Size of keys
	 */
	private int getNewKeySize(){
		return 1+securityParam/8;
	}	
	private int getNewKeysSize(int newKeys){
		return newKeys * getNewKeySize();
	}
	
	
	/**
	 * @param sigKeys
	 * @param blocksCount
	 * @return Size of security tags
	 */
	private int getTagsSize(int sigKeys, int blocksCount, int newKeys){
		return 
			getHeaderSize(sigKeys)
			+
			getHashesSize(blocksCount)
			+
			getNewKeysSize(newKeys)
		;
	}

	/**
	 * Number of blocks required to transfer a file of size dataLength, with blocks of size
	 * blockSize, and a signature of size sigSize. 
	 * Bootstrap block not included.
	 * @param dataLength
	 * @param blockSize
	 * @param sigKeys Number of keys in the signature
	 * @return
	 */
	private int getBlockCount(int dataLength,int blockSize,int sigKeys, int newKeys){
		return 
			(getHeaderSize(sigKeys) + getNewKeysSize(newKeys) + dataLength)
			/
			(blockSize - getHashSize())
			;
	}

	/**
	 * Retrieve the data part of a block.
	 * @param data 
	 * @param dataPadSize
	 * @param blockSize
	 * @param blockIndex     Block 0 is the bootstrap block
	 * @return Data part of the desired block. In this case data contains also new keys.
	 */
	private byte[] getBlockDataPart(byte[] data,int sigKeys,int blockSize,int blockIndex, int newKeys){
		int blocksCount=getBlockCount(data.length, blockSize, sigKeys, newKeys);
		int tagsSize=getTagsSize(sigKeys, blocksCount, newKeys)-getNewKeysSize(newKeys);
/*
        block 0        block 1      block 2       block 3            
	+-------------+-------------+-/--------- /--+-------------+
	|             |   TAGS      | \ New keys \  | DATA        |
	+-------------+-------------+-/----------/--+-------------+
	<---------------------------------------->       
	              tagsSize                     
	              				  <---------->
	              				  newKeysSize
	<-------------------------->               
	   blockIndex * blockSize      (when blockIndex==2)             
	<----------------------------------------->
	        (blockIndex+1) * blockSize (when blockIndex==2)
 */
		int blockStart,blockEnd;		
		blockStart=
			blockIndex*blockSize 
			-
			tagsSize             
		;
		blockEnd=blockStart+blockSize;
		
		if(blockStart<0)blockStart=0;		
		if(blockEnd<0)  blockEnd  =0;
		if(blockStart>data.length)blockStart=data.length;
		if(blockEnd  >data.length)blockEnd  =data.length;
		
		return Arrays.copyOfRange(
				data, 
				blockStart, 
				blockEnd
		);
	}
	
	/**
	 * Retrieve the hashes part of a block.
	 * @param hashes 
	 * @param sigKeys
	 * @param blockSize
	 * @param blockIndex     Block 0 is the bootstrap block
	 * @return Hash part of the desired block.
	 */	
	private byte[] getBlockHashesPart(byte[] hashes,int sigKeys,int blockSize,int blockIndex){
		int headerSize=getHeaderSize(sigKeys);
	
/*
        block 0        block 1      block 2       block 3            
	+--------/----+-------------+-/----------/-+-------------+
	| HEADER \    |     HASHES  | \ NEW KEYS \ |    DATA     |
	+--------/----+-------------+-/----------/-+-------------+
	<-------->                    
	headerSize                    
	          <------------------>       
	              hashes.length                     
	<0------------>1------------>2------------>3------------>4:blockIndex               
	   blockIndex * blockSize                   
	<------------->0------------>1------------>2------------>3:blockIndex
	        (blockIndex+1) * blockSize 
 */
		int blockStart,blockEnd;		
		blockStart=
			blockIndex*blockSize 
			-
			headerSize             
		;
		blockEnd=blockStart+blockSize;
		
		if(blockStart<0)blockStart=0;		
		if(blockEnd<0)  blockEnd  =0;
		if(blockStart>hashes.length)blockStart=hashes.length;
		if(blockEnd  >hashes.length)blockEnd=hashes.length;
		
		return Arrays.copyOfRange(
				hashes, 
				blockStart, 
				blockEnd
		);		
	}
	
	
	
	
	private byte[] getHashes(byte[] data,int blockSize,int sigKeys, byte[] newKeys){				
		byte[] dataAndKeys=new byte[data.length+newKeys.length];
		ArrayUtils.copyInplace(newKeys, dataAndKeys, 0, 0);
		ArrayUtils.copyInplace(data   , dataAndKeys, 0, newKeys.length);
		
		//int blocksCount=getBlockCount(dataAndKeys.length, blockSize, sigKeys, newKeys.length/ getNewKeySize());
		int blocksCount=getBlockCount(data.length, blockSize, sigKeys, newKeys.length/ getNewKeySize());
		byte[] result=new byte[blocksCount*getHashSize()];
//		HashFunction hf=new SHA1HashFunction();		
		
		/*
		  Build hashes starting from the last.
		  We cannot start from the first because the first block can contain hashes of 
		  next blocks, that we have not yet calculated if we start from the first. 
		*/
		for(int i=blocksCount;i>0;i--){
			byte[] block=new byte[blockSize];
			
			byte[] dataPart=getBlockDataPart(
				dataAndKeys, // Because it's the same for calculating hashes
				sigKeys, 
				blockSize, 
				i,
				0 // Because keys are already included in data 
			);
			byte[] hashPart=getBlockHashesPart(
				result, 
				sigKeys, 
				blockSize, 
				i
			);
			// Write in block starting from byte 0 hash part
			ArrayUtils.copyInplace(
				hashPart,
				block,
				0,
				0
			);
			// Write in block starting after hash part data part
			ArrayUtils.copyInplace(
				dataPart,
				block,
				0,
				hashPart.length
			);
			
			block=Arrays.copyOf(block, hashPart.length + dataPart.length);
			
			byte[] hash=hf.calculateHash(
				block,
				getHashSize()
			);
			
			ArrayUtils.copyInplace(
				hash, 
				result, 
				0, 
				(i-1)*hash.length
			);
		}
		return result;
	}
	
	
	/**
	 * Returns unsigned part of security tags. It consists of Session info and hashes of blocks.
	 * @param data
	 * @param blockSize
	 * @param sigKeys
	 * @param hashAdjust
	 * @return
	 * @throws InvalidParameterException
	 */
	private byte[] getUnsignedTags(byte[] data,int blockSize,int sigKeys, byte hashAdjust, byte[] newKeys)throws InvalidParameterException{
		int newKeysCount=newKeys.length/getNewKeySize();
		int blocksCount=getBlockCount(data.length, blockSize, sigKeys, newKeysCount);
				
		//byte[] result=new byte[ST0_SESSION_INFO_SIZE + (blocksCount+1)*getHashSize() + newKeys.length];
		byte[] result=new byte[ST0_SESSION_INFO_SIZE + blocksCount*getHashSize() + newKeys.length];
				
		int offs=0;	

		// Session info section
		// blocks count
		result[offs++]=(byte)blocksCount;
		// new key count
		result[offs++]=(byte)newKeysCount;
		//4 bytes for code image size
		result[offs++]=(byte)(data.length >> 0);
		result[offs++]=(byte)(data.length >> 8);
		result[offs++]=(byte)(data.length >> 16);
		result[offs++]=(byte)(data.length >> 24);
		// Adjust byte
		result[offs++]=hashAdjust;

		// Hashes of blocks
		byte[] hashes=getHashes(data, blockSize, sigKeys, newKeys);		
		ArrayUtils.copyInplace(
			hashes, 
			result, 
			0, 
			offs
		);
		offs+=hashes.length;

		// New keys		
		ArrayUtils.copyInplace(
			newKeys, 
			result, 
			0, 
			offs
		);
		offs+=newKeys.length;		
		
		result=Arrays.copyOf(result,offs);
				
//		System.out.println("Unsigned tags:\n"+Arrays.toString(result));
		
		return result;		
	}
		
	
	/**
	 * Returns signable part of first block, given the expected signature length.
	 * The signable part is made of SessionInfo, hashes of blocks and datapad.
	 * @param data
	 * @param blockSize
	 * @param sigKeys Number of keys in the signature
	 * @param hashAdjust Adjustable byte for signature tuning
	 * @return
	 * @throws InvalidParameterException
	 */
	private byte[] getFirstBlockSignablePart(byte[] data,int blockSize,int sigKeys, byte hashAdjust, byte[] newKeys)throws InvalidParameterException{
		// Available space after signature
		int availableSize=
			blockSize
			-
			getSignatureHeaderSize(sigKeys)
		;		
		byte[] result=new byte[availableSize];
		
		// Get tags and copy 
		byte[] tags;		
		tags=getUnsignedTags(data, blockSize, sigKeys, hashAdjust, newKeys);
		int tagsSize=tags.length;
		if(tagsSize>availableSize)tagsSize=availableSize;
		ArrayUtils.copyInplace(tags, result, 0, 0, tagsSize);
		availableSize-=tagsSize;
		
		// If there is space fill with data
		if(availableSize>0){			
			ArrayUtils.copyInplace(
				data, 
				result, 
				0, 
				tagsSize, 
				availableSize
			);
		}
		
		return result;
	}
	
	
	/**
	 * Contains information about results of a tagging operation.
	 * This information can be used to update keys used in the signature.  
	 * @author Moreno Dissegna
	 */
	public class TaggedData{
		private byte[] originalData,taggedData;
		private TTimeSA.TTimeSASignature signature;
		private TTimeSA.SecretKey[] keysToRefresh; 
		
		public TaggedData(byte[] originalData,byte[] taggedData,TTimeSA.TTimeSASignature signature,TTimeSA.SecretKey[] keysToRefresh){
			this.originalData=originalData;
			this.taggedData=taggedData;
			this.signature=signature;
			this.keysToRefresh=keysToRefresh;
		}		
	}
	
	

	/**
	 * Adds security tags to specified data.
	 * @param data
	 * @param blockSize
	 * @return
	 * @throws Exception
	 */
	private TaggedData createSecurityTags(byte[] data,int blockSize) throws Exception{
		byte[] resultData=null;
		
		// Find appropriate signature length and adjustable byte
		int guessedSigKeys=getStartGuessSigHashes();
		int hashAdjust=0; 
		boolean correctSignature=false;
		byte[] signable=null;
		
		
		TTimeSA.SecretKey[] keysToRefresh=null;
		byte[] newKeys=null;
		
		while(!correctSignature & (guessedSigKeys<=getSigMaxSize())){
			if (useKeyRefresh){
				if(useLowOverheadKeyRefresh)
					keysToRefresh=skr.getKeysToRefresh();
				else
					keysToRefresh=skr.getKeysToRefresh(getSigMaxSize()- guessedSigKeys);
			}else
				keysToRefresh=new TTimeSA.SecretKey[0];
			
			newKeys=new byte[keysToRefresh.length*getNewKeySize()];
			for(int i=0;i<keysToRefresh.length;i++){
				if(isUseLowOverheadKeyRefresh())
					newKeys[i*getNewKeySize()]=(byte)keysToRefresh[i].index;
				else
					newKeys[i*getNewKeySize()]=(byte)i;
				ArrayUtils.copyInplace(keysToRefresh[i].getPublishableKey().value, newKeys, 0, i*getNewKeySize()+1);
			}			
			
			try{
				for(hashAdjust=0;hashAdjust<256;hashAdjust++){
					
						signable=getFirstBlockSignablePart(
								data, 
								blockSize, 
								guessedSigKeys, 
								(byte)hashAdjust,
								newKeys
								);
						int sigKeys=sa.getSignatureLength(skr, signable)/getKeySize();
						if(sigKeys==guessedSigKeys){
							correctSignature=true;
							break;
						}
				}
			}catch(InvalidParameterException e){
				// Invalid dataPadSize, skip this signature length
			}				
			if(!correctSignature)
				guessedSigKeys++;
		}		
		
		if(!correctSignature)
			throw new InvalidParameterException("Impossible to find a valid signature size for this data.");
		
		int blocksCount=getBlockCount(data.length, blockSize, guessedSigKeys, keysToRefresh.length);
		
		resultData=new byte[
		                data.length+
		                getHeaderSize(guessedSigKeys)+
		                getHashesSize(blocksCount)+
		                newKeys.length
		                ];
		
//		System.out.println("Signed message:\n"+Arrays.toString(signable));
		
		System.out.println("Signature length: "+guessedSigKeys+"\nAdjust byte:"+ hashAdjust);
		
		TTimeSA.TTimeSASignature signature=sa.getSignature(skr, signable);
		
		if(signature.value.length!=guessedSigKeys*getKeySize())
			throw new InvalidParameterException("Signature changed...why?");
		
		// Start building result
		int offs=0;
		// Signature size
		resultData[offs++]=(byte)guessedSigKeys;
		// Signature
		ArrayUtils.copyInplace(signature.value, resultData, 0, offs);
		offs+=signature.value.length;
		// Tags
		byte[] tags=getUnsignedTags(data, blockSize, guessedSigKeys, (byte)hashAdjust, newKeys);
		
//		System.out.println("Tags:\n"+Arrays.toString(tags));
		
		ArrayUtils.copyInplace(tags, resultData, 0, offs);
		offs+=tags.length;
		//Data
		ArrayUtils.copyInplace(data, resultData, 0, offs);
		offs+=data.length;			

		if(!verifyTaggedData(resultData, blockSize, skr.getPublishableKeyRing()))
			throw new Exception("Error in signed data verification..why?");
			
		return new TaggedData(data,resultData,signature,keysToRefresh);
	}
	
	public void updateKeys(TaggedData tdata) throws IOException, ParserConfigurationException, SAXException{
		sa.publishedSignature(tdata.signature);
		if(useKeyRefresh && !isUseLowOverheadKeyRefresh()){
			System.out.println("Partial key refreshment index adjust");
			int index=0;
			// Partial key refreshent: key index must be an index over the keys NOT used in the signature
			for(int i=0;i<tdata.signature.usedKeyRing.keyNum;i++){
				if(!tdata.signature.hashBits.get(i)){
					System.out.println("Key index updated from "+tdata.keysToRefresh[index].index+" to "+i);
					tdata.keysToRefresh[index].index=i;
					index++;					
				}
			}			
		}
		
		tdata.signature.usedKeyRing.refreshKeys(tdata.keysToRefresh);
		tdata.signature.usedKeyRing.save();
	}
	
	private boolean verifyTaggedData(byte[] signedData,int blockSize, TTimeSA.PublicKeyRing pkr){	
		int offs=0;
		int sigKeys=(int)signedData[offs++];
		int sigSize=sigKeys*getKeySize();
		byte[] signature=Arrays.copyOfRange(signedData, offs, offs+sigSize);
		offs+=sigSize;
		byte[] signable=Arrays.copyOfRange(signedData, offs, Math.min(blockSize,signedData.length));
		
//		System.out.println(Arrays.toString(signedData));
		
//		System.out.println(Arrays.toString(signable));
		
		// Verify the signature
		if(!sa.verifySignature(pkr, signature, signable))
			return false;		
		
//		System.out.println("Signature valid");
		
		int blockCount,newKeyCount,dataSize;
		blockCount =(int)(signedData[offs++]&0xFF);
		newKeyCount=(int)(signedData[offs++]&0xFF);
		dataSize =(int)(signedData[offs++]<< 0) & 0x000000FF;
		dataSize+=(int)(signedData[offs++]<< 8) & 0x0000FF00;
		dataSize+=(int)(signedData[offs++]<<16) & 0x00FF0000;
		dataSize+=(int)(signedData[offs++]<<24) & 0xFF000000;
		
		System.out.println("Blocks: "+blockCount);
		System.out.println("newKeyCount: "+newKeyCount);
		//System.out.println(dataSize);
		
		// Skip hash adjust byte
		offs++;		
		
//		HashFunction hf=new SHA1HashFunction(); 
		
		//Verify hashes of blocks
		for(int i=1;i<=blockCount;i++){
			int realBlockSize=blockSize;
			if(signedData.length-i*blockSize<blockSize)
				realBlockSize =signedData.length-i*blockSize;
			
			byte[] block=new byte[realBlockSize];
			ArrayUtils.copyInplace(signedData, block, i*blockSize, 0, realBlockSize);
		
			byte[] hash=hf.calculateHash(block, getHashSize());
			byte[] proposedHash=Arrays.copyOfRange(signedData, offs, offs+getHashSize());
			offs+=getHashSize();
			
			//System.out.println("Hash calculated: "+Arrays.toString(hash));
			//System.out.println("Hash proposed  : "+Arrays.toString(proposedHash));			
			
			if(!Arrays.equals(hash, proposedHash))
				return false;
		}
		
		return true;
	}

	
	
	
	
	/**
	 * Command line interface for this class.<br>
	 * Adds security tags to a given file.<br><br>
	 * Command format:<br>
	 *  java SecurityTaggerV0 [-gen-keys &lt;sec.param.&gt; &lt;number of uses&gt; &lt;filename&gt;] [-sign &lt;keys filename&gt; blockSize inputfile outputfile  [-update-keys] [-use-key-refresh [-low-overhead]] ][-get-public-keys &lt;private keys filename&gt; &lt;filename&gt;]
	 *  <br><br>
	 * Options  description:
	 * <ul>
	 * <li>-gen-keys : Generate new private keys with given "security parameter" and "number of uses" and store them in file "filename".
	 * </li>      
	 * <li>-sign : Use keys stored in file "keys filename" and sign "inputfile" , using blocks of size "block size" , and storing output in "outputfile".<br>
	 *  If [-update-keys] is specified, keys are updated, otherwise they are kept unchanged.
	 *  If [-use-key-refresh] is specified, oldest keys are replaced with new ones.
	 *  If [-low-overhead] is specified, low overhead key refreshment is used instead of partial key refreshment.
	 * </li>
	 * <li>-get-public-keys : Stores in file "filename" public keys of private keys in file "private keys filename" 
	 * </ul>
	 * @param args
	 * @throws Exception
	 */
	public static void main(String[] args) throws Exception {
		
		boolean generate=false,sign=false,update=false,getpk=false,keyrefresh=false,useLowOverheadKeyRefresh=false;
		int secParam=0,useCount=0,blockSize=0;
		String newkeysfile="",keysfile="",inputfile="",outputfile="",getpkkeyfile="",pkfile="";
		
		int argi=0;		
		while(argi<args.length){
			if      (args[argi].equals("-gen-keys")){
				generate=true;
				secParam=Integer.parseInt(args[++argi]);
				useCount=Integer.parseInt(args[++argi]);
				newkeysfile=args[++argi];
			}else if(args[argi].equals("-sign")){
				sign=true;
				keysfile=args[++argi];
				blockSize=Integer.parseInt(args[++argi]);
				inputfile=args[++argi];
				outputfile=args[++argi];
			}else if(args[argi].equals("-update-keys")){
				update=true;
			}else if(args[argi].equals("-use-key-refresh")){
				keyrefresh=true;
			}else if(args[argi].equals("-low-overhead")){	
				useLowOverheadKeyRefresh=true;
			}else if(args[argi].equals("-get-public-keys")){
				getpk=true;
				getpkkeyfile=args[++argi];
				pkfile=args[++argi];
			}			
			argi++;
		}
		
		if(args.length==0){
			System.out.println("Usage:\n java SecurityTaggerV0 [-gen-keys <sec.param.> <number of uses> <filename>] [-sign <keys filename> <block size> <inputfile> <outputfile> [-update-keys [-low-overhead]] [-use-key-refresh] ] [-get-public-keys <private keys filename> <filename>]");
			System.out.println("See javadoc for further documentation.");
			System.exit(1);
		}	
		
		SecurityTaggerV0 st0;
		if(generate){
			st0=new SecurityTaggerV0(newkeysfile,secParam,useCount);
			System.out.println("Keys generated.");
		}
		if(sign){
			st0=new SecurityTaggerV0(keysfile);
			
			st0.setUseKeyRefresh(keyrefresh);
			st0.setUseLowOverheadKeyRefresh(useLowOverheadKeyRefresh);
		
			if(!st0.skr.isUsable()){
				System.out.println("Keys expired, impossible to sign.");
				System.out.flush();
				System.exit(1);			
			}else if(st0.skr.expiring()){
				System.out.println("Warning: Some keys are expiring, with this signature they must be refreshed.");
			}
				
			FileInputStream fis=new FileInputStream(inputfile);
			int size=500000;
			int read;
			byte[] data;
			do{
				data=new byte[size];	
				read=fis.read(data);
				if(read==size){
					size*=2;
					fis.close();
					fis=new FileInputStream(inputfile);
				}
			}while(read==size);
			fis.close();
			data=Arrays.copyOf(data, read);			
				
			TaggedData result=st0.createSecurityTags(data, blockSize);
			
			FileOutputStream fos=new FileOutputStream(outputfile);
			fos.write(result.taggedData);
			fos.close();
			
			System.out.println("File signed.");
			
			if(update){
				st0.updateKeys(result);
				//st0.
				System.out.println("Keys updated.");
			}					
		}
		if(getpk){
			TTimeSA sa=new TTimeSA(hf);
			TTimeSA.SecretKeyRing skr=sa.loadKeyRing(getpkkeyfile);
			skr.getPublishableKeyRing().saveToBinary(pkfile);
			
			System.out.println("Public keys retrieved.");
		}
	}
	
}
