/* Copyright (c) 2009 The Regents of the Signet Group.
*  All rights reserved.
*
*  Permission to use, copy, modify, and distribute this software and its
*  documentation for any purpose, without fee, and without written
*  agreement is hereby granted, provided that the above copyright
*  notice, the (updated) modification history and the author appear in
*  all copies of this source code.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS `AS IS'
*  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
*  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
*  ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS
*  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
*  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, LOSS OF USE, DATA,
*  OR PROFITS) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
*  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
*  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
*  THE POSSIBILITY OF SUCH DAMAGE.
*/

package net.tinyos.signet;

import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.security.InvalidKeyException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.BitSet;
import java.util.Comparator;
import java.util.PriorityQueue;
import java.util.Random;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

import com.sun.org.apache.xml.internal.serialize.XMLSerializer;

/**
 * T-Time Signature Algorithm implementation.
 * Based on paper "A ROM-friendly Secure Code Update Mechanism for WSNs 
 *  Using a Stateful-verifier T-time Signature Scheme". 
 *  
 * @author Moreno Dissegna
 */
public class TTimeSA{
	private HashFunction hf;
	private String currentFilename;
		
	public class PublicKey extends Key{

		public PublicKey(int index,byte[] value) {
			super(index,value);
		}

		private boolean validateSecret(Key secretKey, boolean doUpdate){
			boolean result=Arrays.equals(
					hf.calculateHash(secretKey.value, this.value.length),
					this.value
			);
			if(result && doUpdate)
				this.value=secretKey.value.clone();
			return result;
		}
		
		/**
		 * Check key validity and update key if valid
		 * @param secretKey
		 * @return Whether signature is valid 
		 */
		public boolean validateSecret(Key secretKey){
			return validateSecret(secretKey,true);
		}
		/**
		 * Check key validity without updating key 
		 * @param secretKey
		 * @return Whether signature is valid 
		 */
		public boolean validSecret(Key secretKey){
			return validateSecret(secretKey,false);				
		}
		
	}	
	
	public class SecretKey extends Key{
		private int remainingUses;

		/**
		 * Creates a new key initialized with random bytes and given remaining uses count
		 */		
		public SecretKey(int index,int length,int uses) {
			super(index,length);
			remainingUses=uses;
			new Random().nextBytes(value);
		}
		/**
		 * Creates a key with given value and remaining uses count
		 * @param value
		 * @param uses
		 */
		public SecretKey(int index,byte[] value,int uses) {
			super(index,value);
			remainingUses=uses;
		}
		/**
		 * Creates a key with given value and remaining uses count
		 * @param value Value as an hexadecimal string
		 * @param uses
		 */
		public SecretKey(int index,String value,int uses) {
			super(index,value);
			remainingUses=uses;
		}
		
		/**
		 * Check whether key is still usable
		 * @return whether key is still usable
		 */
		public boolean isUsable(){
			return remainingUses>0;
		}

		/**
		 * Get the corresponding publishable key for this secret key.
		 * This key has already been undisclosed.
		 * @return The publishable key
		 * @throws InvalidKeyException
		 */
		public PublicKey getPublishableKey() throws InvalidKeyException{
			if(!isUsable())
				throw new InvalidKeyException("Key expired");
			return new PublicKey(
					this.index,
					hf.calculateHash(
							value, 
							value.length, 
							remainingUses
					)
			);
		}			
		
		/**
		 * Get the corresponding signer key for this secret key.
		 * This key is not yet undisclosed and can be used to sign. 
		 * @return The signer key
		 * @throws InvalidKeyException
		 */
		public PublicKey getSignerKey() throws InvalidKeyException{
			if(!isUsable())
				throw new InvalidKeyException("Key expired");
			return new PublicKey(		
					this.index,
					hf.calculateHash(
							value, 
							value.length, 
							remainingUses-1
					)
			);
		}		
		
		@Override
		public void saveToXML(Element container)
				throws ParserConfigurationException {
			Document doc=container.getOwnerDocument(); 
			super.saveToXML(container);
			Element euses=doc.createElement("remainingUses");
			container.appendChild(euses);
			euses.appendChild(doc.createTextNode(Integer.toString(remainingUses)));
		}
		@Override
		public void loadFromXML(Element container) throws ParserConfigurationException{
			remainingUses=Integer.parseInt(
					container.getElementsByTagName("remainingUses").item(0).getTextContent()
					);
			super.loadFromXML(container);							
		}		
		
		/**
		 * Signals that the publishable key has been used
		 */
		public void published(){
			remainingUses--;
		}
	}
	
	
	private abstract class TTimeSAKeyRing<keyClass extends Key> extends KeyRing<keyClass>{
		protected int keyLength;
		
		public TTimeSAKeyRing(int n) {
			super((n*8) + (int)(Math.log(n*8)/Math.log(2)) + 1);
			this.keyLength=n;			
		}
		
		@Override
		public void saveToXML(Element container) throws ParserConfigurationException{
			Document doc=container.getOwnerDocument();
			super.saveToXML(container);			
			Element elength =doc.createElement("keyLength");
			container.appendChild(elength);
			elength .appendChild(doc.createTextNode(Integer.toString(keyLength)));
		}	
		@Override
		public void loadFromXML(Element container)
				throws ParserConfigurationException {
			super.loadFromXML(container);
			keyLength=Integer.parseInt(
					container.getElementsByTagName("keyLength").item(0).getTextContent()
					);	
		}
	}

	
	/**
	 * Stores a collection of private keys, to be used to sign.
	 * Has methods to load and save to a local file in order to keep status.
	 * @author Moreno Dissegna
	 */
	public class SecretKeyRing extends TTimeSAKeyRing<SecretKey>{
		private int keyValidity;
		
		public SecretKeyRing(int n,int T) {
			super(n);
			generate(T);
		}
		public SecretKeyRing(String filename) throws SAXException, IOException, ParserConfigurationException {
			super(1);
			loadFromFile(filename);
		}		
		
		private void generate(int T){
			keyValidity=T;
			keyList.clear();
			for(int i=0;i<keyNum;i++){
				SecretKey key=new SecretKey(i,keyLength,T);
				keyList.add(key);
			}
		}
		
		public SecretKey[] getKeysToRefresh(){
			/* The number of key to refresh should be N/T, so in T rounds we replace all 
			 * N keys. But this could be done only if we know which keys to replace before 
			 * creating the signature, because we have to sign also the new keys, and this 
			 * cannot be done. So we update the keys that are the oldest BEFORE creating 
			 * the signature. So we have to replace all the keys in T-1 rounds instead of T.
			 */
			int keyCount = this.keyNum / (this.keyValidity -1);
			if(keyCount==0)
				keyCount=1;
			return getKeysToRefresh(keyCount);
		}
		
		public SecretKey[] getKeysToRefresh(int keyCount){			
			SecretKey[] result=new SecretKey[keyCount];
			
			PriorityQueue<SecretKey> q=new PriorityQueue<SecretKey>(this.keyNum,new Comparator<SecretKey>() {			
				@Override
				public int compare(SecretKey o1, SecretKey o2) {
					if(o1.remainingUses==o2.remainingUses)
						return 0;
					else if(o1.remainingUses<o2.remainingUses)
						return -1;
					else
						return 1;					
				}
			});
			
			// put keys in a priority queue ordered on key age
			for (SecretKey	key : this.keyList) 
				q.add(key);				
			
			// find oldest keys and generate replacements
			for(int i=0;i<keyCount;i++){
				SecretKey key=q.poll();
				//System.out.println(key.remainingUses);
				result[i]=new SecretKey(key.index,this.keyLength,this.keyValidity);
			}
					
			// return new keys
			return result;
		}
		
		public void refreshKeys(SecretKey[] newKeys){
			for (SecretKey secretKey : newKeys) {
				System.out.println("Updating key "+secretKey.index+" to value "+Arrays.toString(secretKey.value));
				this.keyList.set(secretKey.index, secretKey);
			}
		}
		
		@Override
		public void saveToXML(Element container) throws ParserConfigurationException{
			Document doc=container.getOwnerDocument();
			super.saveToXML(container);
			Element euses =doc.createElement("keyUses");
			container.appendChild(euses);
			euses .appendChild(doc.createTextNode(Integer.toString(keyValidity)));			
		}		
		
		@Override
		public void loadFromXML(Element container)
				throws ParserConfigurationException {
			super.loadFromXML(container);
			keyValidity=Integer.parseInt(
					container.getElementsByTagName("keyUses").item(0).getTextContent()
					);			
		}
		
		private void loadFromFile(String filename) throws SAXException, IOException, ParserConfigurationException{
			KeyStorage ks=new KeyStorage(filename);			
			loadFromXML(ks.getKeyRingContainer("TTimeSA"));
			
//			Document doc=DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(filename);
//			loadFromXML(doc);
			currentFilename=filename;
		}
			
		public void saveToFile(String filename) throws IOException, ParserConfigurationException, SAXException{
			KeyStorage ks=new KeyStorage(filename);
			ks.addKeyRing("TTimeSA", this);			
			ks.save();
			
//			XMLSerializer xs=new XMLSerializer();
//			xs.setOutputCharStream(new FileWriter(filename));
//			xs.serialize(saveToXML());
			currentFilename=filename;
		}	
		
		public void save() throws IOException, ParserConfigurationException, SAXException{
			if(!currentFilename.equals(""))
				saveToFile(currentFilename);
		}
		

		
		public PublicKeyRing getPublishableKeyRing() throws InvalidKeyException{
			PublicKeyRing result=new PublicKeyRing(keyLength);
			
			for (SecretKey sk : keyList) {
				result.keyList.add(sk.getPublishableKey());
			}
			
			return result;
		}
		
		public PublicKeyRing getSignerKeyRing() throws InvalidKeyException{
			PublicKeyRing result=new PublicKeyRing(keyLength);
			
			for (SecretKey sk : keyList) {
				result.keyList.add(sk.getSignerKey());
			}
			
			return result;
		}		
		
		/**
		 * Use this function to check if there are keys that are going to expire with the next signature.
		 * @return Whether exists any key in the keyring that is going to expire with the next signature.		
		 */
		public boolean expiring(){
			for (SecretKey sk : keyList) {
				if(sk.remainingUses==1)
					return true;
			}		
			return false;
		}
		
		/**
		 * @return Whether the keyring is usable, iff all its keys are usable
		 */
		public boolean isUsable(){
			for (SecretKey sk : keyList) {
				if(!sk.isUsable())
					return false;
			}		
			return true;
		}
		@Override
		protected SecretKey createKey(int index) {
			return new SecretKey(index,new byte[0],0);
		}		
	}
	
	/**
	 * Stores a collection of publishable keys, that can be used to verify a signature.
	 * Has methods to save to a binary file that can be transferred to nodes to allow verification..
	 * @author Moreno Dissegna
	 */
	public class PublicKeyRing extends TTimeSAKeyRing<PublicKey>{
		public PublicKeyRing(int n) {
			super(n);
		}

		/**
		 * Saves keyring to a binary file to be saved on nodes.
		 * File format:
		 * 		byte0: #of keys, less significant bits
		 * 		byte1: #of keys, most significant bits
		 * 		byte2: key size
		 * 		byte3 .. byte(3+ # of keys * key size -1) : array of keys
		 * @return
		 */
		public byte[] saveToBinary(){
//			byte[] result=new byte[3 + keyNum*keyLength];
			byte[] result=new byte[keyNum*keyLength];
			int len=0;
//			result[len++]=(byte)(keyNum    & 0xFF);
//			result[len++]=(byte)(keyNum>>8 & 0xFF);
//			result[len++]=(byte)keyLength;
			for (PublicKey pk : keyList) {
				for(int i=0;i<pk.value.length;i++)
					result[len++]=pk.value[i];
			}
			return result;
		}		
		
		public void saveToBinary(OutputStream os) throws IOException{
			os.write(saveToBinary());
		}
		
		public void saveToBinary(String filename) throws IOException{
			FileOutputStream fos=new FileOutputStream(filename);
			saveToBinary(fos);
		}

		@Override
		protected PublicKey createKey(int index) {
			return new PublicKey(index,new byte[0]);
		}
	}

	
	/**
	 * Keeps information about a signature, in order to split the operations of signature 
	 * generation and key update.
	 * @author Moreno Dissegna
	 */
	public class TTimeSASignature{
		protected byte[] value;
		protected BitSet hashBits;
		protected SecretKeyRing usedKeyRing;
		
		protected TTimeSASignature(byte[] value,BitSet hashBits,SecretKeyRing usedKeyRing) {
			this.value=value.clone();
			this.hashBits=new BitSet();
			this.hashBits.or(hashBits);
			this.usedKeyRing=usedKeyRing;
		}
		/**
		 * @return The signature value
		 */
		public byte[] getValue(){
			return value.clone();
		}
	}
	
	


	/**
	 * Creates an instance of TTimeSA.
	 * @param n (security level in bytes) must be a positive integer
	 * @param T T (number of signature generations/verifications allowed) must be a positive integer
	 */
	public TTimeSA(HashFunction hf){
		this.hf=hf;
	}
	
	/**
	 * Generates a random private keyring
	 * @param keyLength
	 * @param keyUses
	 * @return The newly generated keyring
	 */
	public SecretKeyRing generateNewKeyRing(int keyLength,int keyUses){
		return new SecretKeyRing(keyLength,keyUses);
	}
	/**
	 * Loads a private keyring from a file
	 * @param filename
	 * @return Loaded keyring
	 * @throws SAXException 
	 * @throws IOException  
	 * @throws ParserConfigurationException
	 */
	public SecretKeyRing loadKeyRing(String filename) throws SAXException, IOException, ParserConfigurationException{
		return new SecretKeyRing(filename);
	}	
	
	
	/**
	 * Returns bit representation of hash concatenated with bit representation of zero count
	 */
	private BitSet getBits(byte msg[],int keyLength,int keyCount){
		byte hash[] = hf.calculateHash(msg, keyLength);
		
//		System.out.println("Signed hash:\n"+Arrays.toString(hash));
		
		BitSet hashbs=BitSetUtils.convertToBitSet(hash);
		int zeros=keyLength*8-hashbs.cardinality();
		
		BitSet result=new BitSet(keyCount);
		
		result.clear();
		result.or(hashbs);
				
		
		for(int i=keyLength*8;i<keyCount;i++){
			int pos=keyCount-i-1;
			if((zeros & (1<<pos))!=0)
				result.set(i);
		}	

		return result;
	}

	
	private void copyKey(byte[] signature,int pos,byte[]key){
		for(int i=0;i<key.length;i++)
			signature[i+pos]=key[i];
	}
	private void extractKey(byte[] signature,int pos,byte[]key){
		for(int i=0;i<key.length;i++)
			key[i]=signature[i+pos];
	}	
	
	/**
	 * Returns a preview of the signature length without computing the signature
	 * @param skr
	 * @param msg
	 * @return
	 * @throws InvalidKeyException
	 */
	public int getSignatureLength (SecretKeyRing skr, byte[] msg) throws InvalidKeyException{
		BitSet bs=getBits(msg,skr.keyLength,skr.keyNum);
		return bs.cardinality()*skr.keyLength;
	}
	/**
	 * Calculates the signature using given private keyring
	 * @param skr
	 * @param msg
	 * @return
	 * @throws InvalidKeyException
	 */
	public TTimeSASignature getSignature(SecretKeyRing skr, byte[] msg) throws InvalidKeyException{
		TTimeSASignature result;
		
		byte resultValue[]=new byte[skr.keyNum*skr.keyLength];
		byte pubKey[];
		
		int onesCount=0;
		
		BitSet bs=getBits(msg,skr.keyLength,skr.keyNum);
		for(int i=0;i<bs.length();i++){
			if(bs.get(i)){
				SecretKey sk=skr.getKey(i);
				pubKey=sk.getSignerKey().value;
				
//				System.out.println("Used key "+i+" : "+Arrays.toString(pubKey));
//				byte[] hash=hf.calculateHash(pubKey, skr.keyLength);
//				System.out.println("Hash of key    : "+Arrays.toString(hash));
				
				copyKey(resultValue, onesCount*skr.keyLength, pubKey);
				onesCount++;
			}
		}	
		
		resultValue=Arrays.copyOf(resultValue, onesCount*skr.keyLength);
		
		result=new TTimeSASignature(resultValue,bs,skr);
		
		return result;
	}
	/**
	 * signals that the signature has been published, and so private keyring must be updated.
	 * @param signature
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException 
	 */
	public void publishedSignature(TTimeSASignature signature) throws IOException, ParserConfigurationException, SAXException{
		for(int i=0;i<signature.hashBits.length();i++){
			if(signature.hashBits.get(i)){
				SecretKey sk=signature.usedKeyRing.getKey(i);
				sk.published();
			}
		}	
		signature.usedKeyRing.save();
	}
	
	/**
	 * Verify authenticity of the signature
	 * @param pkr
	 * @param signature
	 * @param msg
	 * @return
	 */
	public boolean verifySignature(PublicKeyRing pkr, byte[] signature,byte[] msg){
		byte hash[];
		
		int onesCount=0;
		
		BitSet bs=getBits(msg,pkr.keyLength,pkr.keyNum );
		byte[] secretKey=new byte[pkr.keyLength];
		
		for(int i=0;i<bs.length();i++){
			if(bs.get(i)){				
				extractKey(signature, onesCount*pkr.keyLength, secretKey);
				Key key=pkr.getKey(i);
				hash=hf.calculateHash(secretKey, pkr.keyLength);
				
				if (!Arrays.equals(hash, key.value))
					return false;
				
				onesCount++;
			}
		}	
				
		return true;
	}
	
	
	
	public static void main(String[] args) throws InvalidKeyException, IOException, ParserConfigurationException, SAXException {
/*		TTimeSA sa=new TTimeSA(new SHA1HashFunction());
		SecretKeyRing askr=sa.generateNewKeyRing(10,30);

		Document doc =DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();
		Element root=doc.createElement("Main");
		doc.appendChild(root);
				
		askr.saveToXML(root);
		
		XMLSerializer xs=new XMLSerializer();
		xs.setOutputCharStream(new OutputStreamWriter(System.out));
		xs.serialize(doc);	
		

/*		
// Test funzionamento getBits
		byte msg[]=new byte[100];
		
		for(int i=0;i<msg.length;i++)
			msg[i]=(byte)i;			
		
		TTimeSA sa=new TTimeSA(10,30,new SHA1HashFunction());
		
		sa.generateKeys();
		
		BitSet bits=sa.getBits(msg);
		
		
		System.out.println(bits.cardinality());
		System.out.println(bits.toString());
		
		for(int i=0;i<sa.getNumOfKeys();i++){
			if (bits.get(i))
				System.out.print("1");
			else
				System.out.print("0");
		}
/**/
/*		
// Test minimo numero uni		
		TTimeSA sa=new TTimeSA(10,30,new SHA1HashFunction());		
		sa.generateKeys();
		
		byte msg[]=new byte[100000];
		BitSet bs;
		
		for(int i=0;i<msg.length;i++)
			msg[i]=(byte)(Math.random()*255);		
		
		int oneCount,minj=0,minOnes=80;
		for(int j=0;j<255;j++){
			msg[0]=(byte)(j);
				
			bs=sa.getBits(msg);
			
			oneCount=0;			
			for(int i=0;i<bs.cardinality();i++)
				oneCount++;
			
			if(oneCount<minOnes){
				minj=j;
				minOnes=oneCount;
			}
		}
		
		msg[0]=(byte)(minj);		
		bs=sa.getBits(msg);	
		
		oneCount=0;			
		for(int i=0;i<bs.cardinality();i++)
			oneCount++;		
		
		System.out.println(bs);
		System.out.println(""+oneCount+" ones");	
		
		byte[] sig=sa.getSignature(msg);
		
		System.out.println(Arrays.toString(sig));
		System.out.println(sig.length);
/**/
		
		
/*

	Uso della classe:
	
	TTimeSA sa=new TTimeSA(new SHA1HashFunction());
	SecretKeyRing skr;
	
	if (key ring ancora da generare){
		skr=sa.generateNewKeyRing(10,30);
		PublicKeyRing pkr=skr.getPublishableKeyRing();
		
		salva skr in locale
		invia pkr sui nodi		
	}else{
		skr.loadFromFile(file locale);		
	}
	
	byte[] msg;
	carica messaggio da firmare in msg
		
	if(
		e' di interesse conoscere in anticipo la lunghezza senza eseguire 
		la firma, per esempio per trovare una firma corta
	){
		int sigLen=getSignatureLength(skr,msg);
	}
		
	TTimeSASignature sig=sa.getSignature(skr,msg);
	
	if(la firma � stata applicata al messaggio con successo){
		sa.publishedSignature(sig);
	}
		
 */
		

		
		
		
		
		
		
		
/*		
		TTimeSA sa=new TTimeSA(new SHA1HashFunction());
		SecretKeyRing skr=sa.generateNewKeyRing(10,30);
		PublicKeyRing pkr=skr.getPublishableKeyRing();
		
		
		skr.saveToFile("./pkr.xml");
		pkr.saveToBinary("nodekeys.binary");
		
		byte msg[]=new byte[100000];
		
		for(int i=0;i<msg.length;i++)
			msg[i]=(byte)(Math.random()*255);		
		
		TTimeSASignature sig=sa.getSignature(skr,msg);
		
		
		
		System.out.println(Arrays.toString(sig.value));
		System.out.println(sig.value.length);
		System.out.println(sa.verifySignature(pkr, sig.value, msg));
		
		sa.publishedSignature(sig);
		
		skr=sa.loadKeyRing("./pkr.xml");
		pkr=skr.getPublishableKeyRing();
		
		sig=sa.getSignature(skr,msg);		
				
		
		System.out.println(Arrays.toString(sig.value));
		System.out.println(sig.value.length);
		System.out.println(sa.verifySignature(pkr, sig.value, msg));		
*/		
	}



	
}
